@@ -15,6 +15,8 @@
 from .llama import *
 import os
 from ._utils import __version__
+from unsloth_zoo.utils import _get_dtype
+from unsloth_zoo.hf_utils import dtype_from_config
 from .llama import (
     LlamaRotaryEmbedding,
     LlamaLinearScalingRotaryEmbedding,
@@ -230,7 +232,7 @@ def MistralForCausalLM_fast_forward(
                     attention_mask = attention_mask.expand(bsz, 1, q_len, q_len)
                 attention_mask = attention_mask + causal_mask_values[None, None, :, :]
 
-            attention_mask = attention_mask.to(dtype=_get_dtype(self.config.torch_dtype))
+            attention_mask = attention_mask.to(dtype=_get_dtype(dtype_from_config(self.config)))
 
     output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
     output_hidden_states = (
@@ -298,17 +300,30 @@ def MistralForCausalLM_fast_forward(
         # < 1024 Normal Unsloth uses less VRAM!
         if bsz * q_len <= 1024: RETURN_LOGITS = True
 
-        if not RETURN_LOGITS and HAS_CUT_CROSS_ENTROPY and labels is not None:
+        if not RETURN_LOGITS and labels is not None:
             n_items = kwargs.get("num_items_in_batch", None) or kwargs.get("n_items", None)
             logit_softcapping = getattr(self.config, "final_logit_softcapping", 0)
-            loss = fused_linear_cross_entropy(
-                hidden_states = hidden_states,
-                lm_weight = lm_head,
-                labels = labels,
-                num_items_in_batch = n_items,
-                logit_softcapping = logit_softcapping,
-            )
 
+            # loss = fused_linear_cross_entropy(
+            #     hidden_states = hidden_states,
+            #     lm_weight = lm_head,
+            #     labels = labels,
+            #     num_items_in_batch = n_items,
+            #     logit_softcapping = logit_softcapping,
+            # )
+            loss = unsloth_fused_ce_loss(
+                trainer              = None,
+                hidden_states        = hidden_states,
+                lm_head_weight       = lm_head,
+                lm_head_bias         = None,
+                labels               = labels,
+                mask                 = None,
+                n_items              = n_items,
+                scaling              = getattr(self, "accelerator_scaler", None),
+                target_gb            = None,
+                torch_compile        = True,
+                logit_softcapping    = logit_softcapping,
+            )
             if not return_dict:
                 output = (logits,) + outputs[1:]
                 return (loss,) + output if loss is not None else output
@@ -324,7 +339,7 @@ def MistralForCausalLM_fast_forward(
         pass
         logits = self.lm_head(hidden_states.to(lm_head.dtype))
     pass
-    logits = logits.to(_get_dtype(self.config.torch_dtype))
+    logits = logits.to(_get_dtype(dtype_from_config(self.config)))
 
     loss = None
     if labels is not None:
