@@ -28,7 +28,7 @@ pass
 from ..kernels import (
     post_patch_loss_function,
 )
-from ._utils import __version__
+from ._utils import __version__, importlib_version
 from ._utils import *
 from ..save import patch_saving_functions
 from peft import LoraConfig, TaskType, get_peft_model as _get_peft_model
@@ -43,8 +43,18 @@ from transformers.models.llama.modeling_llama import logger
 from transformers import __version__ as transformers_version
 from triton import __version__ as triton_version
 from unsloth_zoo.utils import _get_dtype
+from unsloth_zoo.hf_utils import (
+    dtype_from_config,
+    add_dtype_kwargs,
+    fix_lora_auto_mapping,
+    get_auto_processor,
+)
 from unsloth_zoo.patching_utils import patch_model_and_tokenizer
 from unsloth_zoo.training_utils import prepare_model_for_training
+
+from unsloth_zoo.utils import Version
+from transformers import __version__ as transformers_version
+
 import types
 import functools
 import os
@@ -69,10 +79,24 @@ __all__ = [
 
 global NUM_LOGITS_TO_KEEP
 NUM_LOGITS_TO_KEEP = dict()
-global PROMPT_LOOPKUP
-PROMPT_LOOPKUP = dict()
+
+VLLM_SUPPORTED_VLM = [
+    "qwen2_5_vl",
+    "gemma3",
+    "mistral3",
+]
+VLLM_NON_LORA_VLM = [
+    "mllama",
+]
+PRE_COMPILE_INFERENCE = [
+    "gpt_oss",
+]
+
+from transformers import GenerationConfig, CompileConfig, HybridCache, AutoConfig, PretrainedConfig
+HAS_TORCH_DTYPE = "torch_dtype" in PretrainedConfig.__doc__
 
 from transformers import GenerationConfig, CompileConfig, HybridCache
+
 _compile_config = CompileConfig(
     fullgraph = False,
     dynamic = None,
@@ -118,7 +142,7 @@ def unsloth_base_fast_generate(
     bsz = input_ids.shape[0]
 
     FastBaseModel.for_inference(self)
-    dtype = _get_dtype(self.config.torch_dtype)
+    dtype = _get_dtype(dtype_from_config(self.config))
 
     # Check if VLM
     is_vlm = any(
@@ -160,15 +184,6 @@ def unsloth_base_fast_generate(
     key = NUM_LOGITS_TO_KEEP[arch]
     if key is not None and key not in kwargs:
         kwargs[key] = 1
-    global PROMPT_LOOPKUP
-    if arch not in PROMPT_LOOPKUP:
-        # Only works for VLMs and not LLMs!
-        if is_vlm:
-            PROMPT_LOOPKUP[arch] = False
-        else:
-            PROMPT_LOOPKUP[arch] = True
-    if bsz == 1 and PROMPT_LOOPKUP[arch]:
-        kwargs["prompt_lookup_num_tokens"] = 3
 
     # Check pad_token
     model_eos_token_id = getattr(self.config, "eos_token_id", None)
@@ -204,19 +219,26 @@ def unsloth_base_fast_generate(
     # Fix generation_config
     # Use hybrid if sliding window seen, otherwise try static
     cache_implementation = getattr(self.config, "cache_implementation", None)
-    if getattr(self, "_supports_static_cache", True):
+    if getattr(self, "_supports_static_cache", getattr(self, "_can_compile_fullgraph", True)):
         if os.environ.get("UNSLOTH_DISABLE_STATIC_GENERATION", "0") == "0":
             cache_implementation = "static"
-        else:
+        elif Version(transformers_version) < Version("4.56.0.dev0"):
             cache_implementation = None
+        else:
+            # Should work in latest transformers!
+            cache_implementation = "static"
     else:
         cache_implementation = None
     if cache_implementation is not None:
         swa = getattr(getattr(self.config, "text_config", self.config), "sliding_window", None)
-        if swa == 0 or type(swa) is not int:
+        if (swa == 0 or type(swa) is not int) \
+            and (getattr(self, "_can_compile_fullgraph", True) is True):
             cache_implementation = "static"
         else:
-            cache_implementation = "hybrid"
+            if Version(transformers_version) < Version("4.56.0.dev0"):
+                cache_implementation = "hybrid"
+            else:
+                cache_implementation = "static"
 
     if "generation_config" in kwargs:
         kwargs["generation_config"].cache_implementation = cache_implementation
@@ -228,24 +250,36 @@ def unsloth_base_fast_generate(
             kwargs["compile_config"] = _compile_config
     pass
 
-    try:
-        with torch.inference_mode(), autocaster:
-            output = self._old_generate(*args, **kwargs)
-    except:
-        PROMPT_LOOPKUP[arch] = False
-        kwargs.pop("prompt_lookup_num_tokens", None)
-        with torch.inference_mode(), autocaster:
-            output = self._old_generate(*args, **kwargs)
-    finally:
-        pass
-        # return_lora_modules(self, state_dict, torch.float32)
+    # Delete cached Flex Attention masks to reset inference
+    for name, module in self.named_modules():
+        if hasattr(module, "_flex_attention_cache"):
+            try: del module._flex_attention_cache
+            except: pass
+        # Solves AttributeError: 'SlidingWindowLayer' object has no attribute 'max_batch_size'
+        if hasattr(module, "_cache") and "cache_utils" in str(module._cache.__class__):
+            try: del module._cache
+            except: pass
     pass
 
-    FastBaseModel.for_training(self)
+    # DO INFERENCE
+    with torch.inference_mode(), autocaster:
+        output = self._old_generate(*args, **kwargs)
+
+    # Delete cached Flex Attention masks to reset inference
+    for name, module in self.named_modules():
+        if hasattr(module, "_flex_attention_cache"):
+            try: del module._flex_attention_cache
+            except: pass
+        # Solves AttributeError: 'SlidingWindowLayer' object has no attribute 'max_batch_size'
+        if hasattr(module, "_cache") and "cache_utils" in str(module._cache.__class__):
+            try: del module._cache
+            except: pass
+    pass
+
+    # FastBaseModel.for_training(self)
     return output
 pass
 
-
 class FastBaseModel:
 
     @staticmethod
@@ -255,6 +289,7 @@ class FastBaseModel:
         dtype             = None,
         load_in_4bit      = True,
         load_in_8bit      = False,
+        load_in_16bit     = False,
         full_finetuning   = False,
         token             = None,
         device_map        = "sequential",
@@ -266,8 +301,22 @@ class FastBaseModel:
         supports_sdpa     = True,
         whisper_language  = None,
         whisper_task      = None,
+        auto_config       = None,
+        offload_embedding = False,
+        # vLLM parameters
+        fast_inference    = False,
+        gpu_memory_utilization = 0.5,
+        float8_kv_cache   = False,
+        random_state      = 3407,
+        max_lora_rank     = 64,
+        disable_log_stats = False,
+        unsloth_vllm_standby = False,
         **kwargs,
     ):
+        if unsloth_vllm_standby and os.environ.get("UNSLOTH_VLLM_STANDBY", "0") != "1":
+            raise RuntimeError("Unsloth: UNSLOTH_VLLM_STANDBY is True, but UNSLOTH_VLLM_STANDBY is not set to 1!")
+        pass
+
         if model_types is None:
             raise RuntimeError(
                 "Unsloth: Please use FastModel or FastVisionModel and not use FastBaseModel directly!"
@@ -275,6 +324,31 @@ class FastBaseModel:
         if os.environ.get("UNSLOTH_MODEL_NAME", "") == "":
             os.environ["UNSLOTH_MODEL_NAME"] = model_name.lower()
 
+        is_vlm = (auto_model in [AutoModelForVision2Seq, AutoModelForImageTextToText])
+        is_whisper = (whisper_language is not None and whisper_task is not None)
+        auto_processor = AutoProcessor if (is_vlm or is_whisper) else AutoTokenizer
+
+        model_type_arch = model_types[0]
+        if model_type_arch == "siglip":
+            for model_type_arch in model_types:
+                if model_type_arch != "siglip": break
+
+        vllm_enable_lora = True
+
+        if is_vlm and fast_inference:
+            if not any(arch in VLLM_SUPPORTED_VLM for arch in model_types):
+                raise RuntimeError(
+                    f"Unsloth: Fast inference is only supported for Language models and Qwen2.5-VL, Gemma3 among vision models. "
+                    f"Found architectures: {', '.join(model_types)}!"
+                )
+
+        if any(arch in VLLM_NON_LORA_VLM for arch in model_types):
+            # mllama is still only in vllm v0 https://arc.net/l/quote/llwkfgmu
+            # https://docs.vllm.ai/en/stable/models/supported_models.html#text-generation_1
+            # vLLM V0 does not support LoRA on multi modal models.
+            # TODO: Update this once vLLM V1 supports Llama 3.2 aka mllama
+            vllm_enable_lora = False
+
         os.environ["UNSLOTH_USE_NEW_MODEL"] = "1"
         if trust_remote_code:
             print(
@@ -289,15 +363,18 @@ class FastBaseModel:
             gpu_stats = torch.cuda.get_device_properties(0)
             gpu_version = torch.version.cuda
             gpu_stats_snippet = f"CUDA: {gpu_stats.major}.{gpu_stats.minor}. CUDA Toolkit: {gpu_version}."
-
-            from importlib.metadata import version as importlib_version
+            try:    vllm_version = f" vLLM: {importlib_version('vllm')}."
+            except: vllm_version = ""
+        elif DEVICE_TYPE == "hip":
+            gpu_stats = torch.cuda.get_device_properties(0)
+            gpu_version = torch.version.hip
+            gpu_stats_snippet = f"ROCm Toolkit: {gpu_version}."
             try:    vllm_version = f" vLLM: {importlib_version('vllm')}."
             except: vllm_version = ""
         elif DEVICE_TYPE == "xpu":
             gpu_stats = torch.xpu.get_device_properties(0)
             gpu_version = torch.version.xpu
             gpu_stats_snippet = f"Intel Toolkit: {gpu_version}."
-
             # TODO: After adding vLLM support for XPU, changed this
             vllm_version = ""
         else:
@@ -305,11 +382,6 @@ class FastBaseModel:
 
         max_memory = round(gpu_stats.total_memory / 1024 / 1024 / 1024, 3)
 
-        model_type_arch = model_types[0]
-        if model_type_arch == "siglip":
-            for model_type_arch in model_types:
-                if model_type_arch != "siglip": break
-
         statistics = \
         f"==((====))==  Unsloth {__version__}: Fast {model_type_arch.title()} patching. Transformers: {transformers_version}.{vllm_version}\n"\
         f"   {chr(92)}{chr(92)}   /|    {gpu_stats.name}. Num GPUs = {DEVICE_COUNT}. Max memory: {max_memory} GB. Platform: {platform_system}.\n"\
@@ -358,12 +430,13 @@ class FastBaseModel:
             custom_datatype = os.environ["UNSLOTH_FORCE_CUSTOM_DTYPE"]
             assert custom_datatype.count(";") >= 4
             checker, _dtype, _bnb_compute_dtype, _custom_datatype, execute_code = custom_datatype.split(";", 4)
-
             # Allow custom dtypes on all runs
             allow_all_runs = (checker == "all")
             # Allow only on float16 datatypes
-            allow_float16_runs = (checker == "float16" and dtype == torch.float16)
-
+            allow_float16_runs = (
+                (checker == "float16" or checker == "torch.float16") and \
+                (dtype == torch.float16 or os.environ.get("UNSLOTH_FORCE_FLOAT32", "0") == "1")
+            )
             if allow_all_runs or allow_float16_runs:
                 if eval(_dtype) is not None:
                     dtype = eval(_dtype)
@@ -383,19 +456,21 @@ class FastBaseModel:
         if not ("attn_implementation" in kwargs):
             kwargs["attn_implementation"] = "sdpa"
         if not supports_sdpa:
-            print(f"Unsloth: {model_type_arch.title()} does not support SDPA - switching to eager!")
+            if os.environ.get("UNSLOTH_ENABLE_FLEX_ATTENTION", "0") == "0":
+                print(f"Unsloth: {model_type_arch.title()} does not support SDPA - switching to fast eager.")
             del kwargs["attn_implementation"]
         pass
 
         bnb_config = None
         if full_finetuning and (load_in_4bit or load_in_8bit):
             print("Unsloth: You selected full finetuning support, but 4bit / 8bit is enabled - disabling LoRA / QLoRA.")
-            load_in_4bit = False
-            load_in_8bit = False
+            load_in_4bit  = False
+            load_in_8bit  = False
+            load_in_16bit = False
         pass
 
-        if load_in_4bit and load_in_8bit:
-            raise RuntimeError("Unsloth: Can only load in 4bit or 8bit, not both!")
+        if int(load_in_4bit) + int(load_in_8bit) + int(load_in_16bit) >= 2:
+            raise RuntimeError("Unsloth: Can only load in 4bit or 8bit or 16bit, not a combination!")
         if load_in_4bit:
             bnb_config = BitsAndBytesConfig(
                 load_in_4bit              = True,
@@ -409,6 +484,8 @@ class FastBaseModel:
                 load_in_8bit              = True,
                 llm_int8_skip_modules     = SKIP_QUANTIZATION_MODULES.copy(),
             )
+        elif load_in_16bit:
+            bnb_config = None
         elif not load_in_4bit and not load_in_8bit and not full_finetuning:
             print("Unsloth: QLoRA and full finetuning all not selected. Switching to 16bit LoRA.")
         pass
@@ -430,31 +507,129 @@ class FastBaseModel:
         # Cannot be None, since HF now checks for the config
         if load_in_4bit:
             # Ignore load_in_4bit / load_in_8bit for MXFP4 - best to get config file
-            if "gpt-oss" in model_name.lower():
+            if "gpt-oss-20b" in model_name.lower() or "gpt-oss-120b" in model_name.lower():
                 pass
             else:
                 kwargs["quantization_config"] = bnb_config
+        else:
+            if auto_config is None:
+                auto_config = AutoConfig.from_pretrained(
+                    model_name,
+                    token = token,
+                    trust_remote_code = trust_remote_code,
+                )
+            if hasattr(auto_config, "quantization_config"):
+                from transformers.quantizers.auto import AUTO_QUANTIZATION_CONFIG_MAPPING
+                quantization_config = auto_config.quantization_config
+                quantizer = AUTO_QUANTIZATION_CONFIG_MAPPING[quantization_config["quant_method"]]
+                quantizer_kwargs = {}
+                # We cannot dequantize since gpt-oss-20b MXFP4 will now be gpt-oss-20b-BF16
+                # if "dequantize" in inspect.signature(quantizer).parameters:
+                #     quantizer_kwargs["dequantize"] = True
+                quantization_config = quantizer.from_dict(quantization_config, **quantizer_kwargs)
+                kwargs["quantization_config"] = quantization_config
+            pass
         pass
 
         # Check if using forced float32 - we load it in bfloat16, then cast to float16!
         torch_dtype = dtype
         if do_forced_float32: torch_dtype = torch.bfloat16
 
+        kwargs = add_dtype_kwargs(torch_dtype, kwargs)
+
         raise_handler = RaiseUninitialized()
-        model = auto_model.from_pretrained(
-            model_name,
-            device_map              = device_map,
-            torch_dtype             = torch_dtype,
-            # quantization_config   = bnb_config,
-            token                   = token,
-            trust_remote_code       = trust_remote_code,
-            # attn_implementation   = attn_implementation,
-            **kwargs,
-        )
+        if not fast_inference:
+            model = auto_model.from_pretrained(
+                model_name,
+                device_map              = device_map,
+                # torch_dtype           = torch_dtype, # Transformers removed torch_dtype
+                # quantization_config   = bnb_config,
+                token                   = token,
+                trust_remote_code       = trust_remote_code,
+                # attn_implementation   = attn_implementation,
+                **kwargs,
+            )
+            if hasattr(model, 'generate'):
+                model.fast_generate = model.generate
+                model.fast_generate_batches = error_out_no_vllm
+            if offload_embedding:
+                embed_tokens = model.get_input_embeddings()
+                nbytes = embed_tokens.weight.numel() * embed_tokens.weight.itemsize
+                ngb = round(nbytes / 1024 / 1024 / 1024, 2)
+                print(f"Unsloth: Offloading embeddings to RAM to save {ngb} GB.")
+                embed_tokens.to("cpu")
+
+                # Add hooks to move inputs to CPU and back to CUDA
+                # [TODO] Doesn't seem to work!
+                # def pre_hook(module, args):
+                #     args[0]._old_device = args[0].device
+                #     return (args[0].to("cpu", non_blocking = True))
+                # def post_hook(module, args, output):
+                #     old_device = getattr(args[0], "_old_device", "cuda")
+                #     return output.to(old_device, non_blocking = True)
+                # embed_tokens.register_forward_pre_hook(pre_hook,  prepend = True)
+                # embed_tokens.register_forward_hook    (post_hook, prepend = True)
+                # Must free GPU memory otherwise will not free!
+                torch.cuda.empty_cache()
+                gc.collect()
+        else:
+            from unsloth_zoo.vllm_utils import (
+                load_vllm,
+                get_vllm_state_dict,
+                convert_vllm_to_huggingface,
+                generate_batches,
+            )
+            model_config = AutoConfig.from_pretrained(
+                model_name,
+                token = token,
+                attn_implementation = "sdpa" if supports_sdpa else "eager",
+            )
+
+            if fast_inference:
+                fast_inference, model_name = fast_inference_setup(model_name, model_config)
+
+            allowed_args = inspect.getfullargspec(load_vllm).args
+            load_vllm_kwargs = dict(
+                model_name             = model_name,
+                config                 = model_config,
+                gpu_memory_utilization = gpu_memory_utilization,
+                max_seq_length         = max_seq_length,
+                dtype                  = dtype,
+                float8_kv_cache        = float8_kv_cache,
+                enable_lora            = vllm_enable_lora,
+                max_lora_rank          = max_lora_rank,
+                disable_log_stats      = disable_log_stats,
+                use_bitsandbytes       = load_in_4bit,
+                unsloth_vllm_standby   = unsloth_vllm_standby,
+                is_vision_model        = is_vlm,
+            )
+            for allowed_arg in allowed_args:
+                if allowed_arg not in load_vllm_kwargs and allowed_arg in kwargs:
+                    load_vllm_kwargs[allowed_arg] = kwargs[allowed_arg]
+            pass
+
+            # Load vLLM first
+            llm = load_vllm(**load_vllm_kwargs)
+
+            # Convert to HF format
+            _, quant_state_dict = get_vllm_state_dict(llm, config = model_config, is_vision_model = True)
+            model = convert_vllm_to_huggingface(quant_state_dict, model_config, dtype, bnb_config, is_vision_model = True)
+            model.vllm_engine = llm
+            model.fast_generate = model.vllm_engine.generate
+            model.fast_generate_batches = functools.partial(generate_batches, model.vllm_engine)
+        pass
+
         raise_handler.remove()
+
         # Return old flag
         os.environ["HF_HUB_ENABLE_HF_TRANSFER"] = old_hf_transfer
 
+        # Check float32 norm weights
+        if os.environ.get("UNSLOTH_HIGH_PRECISION_LAYERNORM", "0") == "1":
+            for jj, (name, module) in enumerate(model.named_modules()):
+                if name.endswith("norm") and hasattr(module, "weight"):
+                    module._pre_set_compute_dtype = torch.float32
+        pass
         # Edit data-types
         if custom_datatype is not None:
             with torch.no_grad():
@@ -466,33 +641,37 @@ class FastBaseModel:
         # Clear deleted GPU items
         for _ in range(3):
             gc.collect()
-            if DEVICE_TYPE == "cuda":  torch.cuda.empty_cache()
+            if DEVICE_TYPE in ("cuda", "hip"):  torch.cuda.empty_cache()
             elif DEVICE_TYPE == "xpu": torch.xpu.empty_cache()
         pass
 
         # Counteract saved tokenizers
         tokenizer_name = model_name if tokenizer_name is None else tokenizer_name
-        is_vlm = (auto_model is AutoModelForVision2Seq)
-        is_whisper = (whisper_language is not None and whisper_task is not None)
-        auto_processor = AutoProcessor if (is_vlm or is_whisper) else AutoTokenizer
         if (whisper_language and whisper_task) or auto_model.__name__.endswith("ForConditionalGeneration"):
            tokenizer = auto_processor.from_pretrained(
                 tokenizer_name,
-                padding_side = "right",
+                padding_side = "left",
                 token        = token,
                 language     = whisper_language,
                 task         = whisper_task,
             )
         else:
-            tokenizer = auto_processor.from_pretrained(
-                tokenizer_name,
-                padding_side = "right",
-                token        = token,
-            )
+            try:
+                tokenizer = auto_processor.from_pretrained(
+                    tokenizer_name,
+                    padding_side = "left",
+                    token        = token,
+                )
+            except:
+                tokenizer = get_auto_processor(
+                    tokenizer_name,
+                    padding_side = "left",
+                    token        = token,
+                )
         if hasattr(tokenizer, "tokenizer"):
             __tokenizer = tokenizer.tokenizer
             # Add padding side as well
-            __tokenizer.padding_side = "right"
+            __tokenizer.padding_side = "left"
             # Check bos, eos, pad tokens
             if hasattr(__tokenizer, "bos_token"):
                 tokenizer.bos_token    = __tokenizer.bos_token
@@ -540,12 +719,15 @@ class FastBaseModel:
             m = m.model
         pass
         m.max_seq_length = max_seq_length
+        # Save to modules as well
+        for module in model.modules():
+            module.max_seq_length = max_seq_length
         m._saved_temp_tokenizer = tokenizer
         # Also set is_loaded_in_8bit to disable incorrect DDP
         m.is_loaded_in_8bit = True if not full_finetuning else False
 
         # Patch generate
-        if os.environ.get("UNSLOTH_DISABLE_FAST_GENERATION", "0") == "0":
+        if os.environ.get("UNSLOTH_DISABLE_FAST_GENERATION", "0") == "0" and hasattr(model, 'generate'):
             if model.generate.__name__ != "unsloth_base_fast_generate":
                 model._old_generate = model.generate
                 unsloth_base_fast_generate.__doc__ = model._old_generate.__doc__
@@ -557,11 +739,13 @@ class FastBaseModel:
             model,
             use_gradient_checkpointing = use_gradient_checkpointing,
             trust_remote_code  = trust_remote_code,
+            model_type = model_type_arch,
+            tokenizer = tokenizer,
         )
         # Clear deleted GPU items
         for _ in range(3):
             gc.collect()
-            if DEVICE_TYPE == "cuda":
+            if DEVICE_TYPE in ("cuda", "hip"):
                 torch.cuda.empty_cache()
             elif DEVICE_TYPE == "xpu":
                 torch.xpu.empty_cache()
@@ -569,6 +753,51 @@ class FastBaseModel:
         return model, tokenizer
     pass
 
+    @staticmethod
+    def pre_compile_for_inference(model_type, model, tokenizer):
+        """
+        We need to invoke torch.compile to save VRAM usage and make it faster downstream.
+        Sometimes torch.compile can use 3GB weirdly on large batches, then it goes down to <1GB.
+        So we invoke torch.compile on short batches to reduce VRAM usage.
+        """
+        if model_type is None or model is None or tokenizer is None: return
+        if str(model_type).lower() not in PRE_COMPILE_INFERENCE: return
+        if getattr(tokenizer, "chat_template", None) is None: return
+        # Check if already compiled and exit
+        for module in model.modules():
+            if hasattr(module, "_pre_compiled_for_inference"): return
+        pass
+        print(f"🦥 Unsloth: Pre compiling {model_type.title()} model for faster inference - this might take 3 minutes or so!")
+        print("========= Pre compiling model for faster inference. Please be patient thank you! =========")
+        # Do single inference
+        messages = [
+            [
+                 {"role": "user", "content": f"What is 1+1 equal to?"},
+            ],
+        ]*1
+        inputs = tokenizer.apply_chat_template(
+            messages,
+            add_generation_prompt = True,
+            return_tensors = "pt",
+            return_dict = True,
+        ).to(model.device)
+        _ = model.generate(**inputs, max_new_tokens = 1)
+        # Do batched inference
+        messages = [
+            [
+                 {"role": "user", "content": f"1+1"},
+            ],
+        ]*4
+        inputs = tokenizer.apply_chat_template(
+            messages,
+            add_generation_prompt = True,
+            return_tensors = "pt",
+            return_dict = True,
+        ).to(model.device)
+        _ = model.generate(**inputs, max_new_tokens = 2)
+        # Set we already pre compiled
+        model._pre_compiled_for_inference = True
+    pass
 
     @staticmethod
     def get_peft_model(
@@ -627,55 +856,76 @@ class FastBaseModel:
             assert(type(target_modules) in (list, tuple, str,))
         pass
 
+        if hasattr(model, "vllm_engine"):
+            if hasattr(model.vllm_engine, "llm_engine") and hasattr(model.vllm_engine.llm_engine, "vllm_config") and getattr(model.vllm_engine.llm_engine.vllm_config, "lora_config", None) is None:
+                # If vLLM is being used but lora is not enabled, throw an error
+                # Ref https://github.com/vllm-project/vllm/blob/51ba839555a5d122eadd91e9c16463ac288f5fa1/vllm/v1/engine/processor.py#L148-L151
+                raise RuntimeError("Unsloth: LoRA is not enabled for this model!")
+            if finetune_vision_layers:
+                # vLLM does not support LoRA on vision layers
+                # https://github.com/vllm-project/vllm/blob/main/vllm/lora/models.py#L471-L477
+                # TODO: Update this once vLLM V1 supports LoRA on vision layers (possibly not happening)
+                raise RuntimeError("Unsloth: Finetuning vision layers is not supported for fast_inference. Only text layers are supported!")
+            if model.config.model_type in VLLM_NON_LORA_VLM:
+                # mllama is still only in vllm v0 https://arc.net/l/quote/llwkfgmu
+                # https://docs.vllm.ai/en/stable/models/supported_models.html#text-generation_1
+                # vLLM V0 does not support LoRA on multi modal models.
+                # TODO: Update this once vLLM V1 supports Llama 3.2 aka mllama
+                raise RuntimeError("Unsloth: LoRA finetuning for Llama 3.2 aka mllama models is not supported with fast_inference!")
+
         # Clear deleted GPU items
         for _ in range(3):
             gc.collect()
-            if DEVICE_TYPE == "cuda":
+            if DEVICE_TYPE in ("cuda", "hip"):
                 torch.cuda.empty_cache()
             elif DEVICE_TYPE == "xpu":
                 torch.xpu.empty_cache()
         pass
         max_seq_length = model.max_seq_length
-        # if we pass loftq_config = None we will get an error
+        # If we pass loftq_config = None we will get an error
         loftq_config = validate_loftq_config(loftq_config, lora_dropout, bias, init_lora_weights, model)
-        lora_config_dict = {
-            "r"                 : r,
-            "lora_alpha"        : lora_alpha,
-            "target_modules"    : target_modules,
-            "target_parameters" : kwargs.get("target_parameters", None),
-            "lora_dropout"      : lora_dropout,
-            "bias"              : bias,
-            "task_type"         : task_type,
-            "use_rslora"        : use_rslora,
-            "init_lora_weights" : init_lora_weights,
-            "loftq_config"      : loftq_config,
-        }
+
+        # Get only allowed parameters for LoraConfig
+        local_variables = { **locals(), **kwargs, }
+        del local_variables["kwargs"]
+        allowed_parameters = inspect.signature(LoraConfig).parameters.keys()
         lora_config = LoraConfig(
-            **{k:v for k,v in lora_config_dict.items() if k in LoraConfig.__doc__},
+            **{ k : v for k, v in local_variables.items() if k in allowed_parameters },
         )
         model = prepare_model_for_kbit_training(
             model,
             use_gradient_checkpointing = use_gradient_checkpointing,
         )
         model = _get_peft_model(model, lora_config)
+        # Fix LoraConfig.auto_mapping is None
+        fix_lora_auto_mapping(model)
         # Enable gradients on modules which are trainable
         requires_grad_for_gradient_checkpointing(model)
         trust_remote_code = getattr(model, "_unsloth_trust_remote_code", False)
         model = FastBaseModel.post_patch_model(model, use_gradient_checkpointing, trust_remote_code = trust_remote_code)
         model.max_seq_length = max_seq_length
+        # Save to modules as well
+        for module in model.modules():
+            module.max_seq_length = max_seq_length
         # Clear deleted GPU items
         for _ in range(3):
             gc.collect()
-            if DEVICE_TYPE == "cuda":
+            if DEVICE_TYPE in ("cuda", "hip"):
                 torch.cuda.empty_cache()
             elif DEVICE_TYPE == "xpu":
                 torch.xpu.empty_cache()
         pass
         patch_saving_functions(model, vision = True)
+        patch_peft_fast_inference(model)
 
         # Add for_inference and for_training
         model.for_training  = functools.partial(FastBaseModel.for_training,  model)
         model.for_inference = functools.partial(FastBaseModel.for_inference, model)
+        m = model
+        while hasattr(m, "model"):
+            m.for_training  = functools.partial(FastBaseModel.for_training,  m)
+            m.for_inference = functools.partial(FastBaseModel.for_inference, m)
+            m = m.model
         return model
     pass
 
@@ -685,11 +935,13 @@ class FastBaseModel:
         model,
         use_gradient_checkpointing = True,
         trust_remote_code = False,
+        model_type = None,
+        tokenizer = None,
     ):
         full_finetuning = os.environ.get("UNSLOTH_ENABLE_FULL_FINETUNING", "0") == "1"
 
         float32_mixed_precision = True
-        if _get_dtype(model.config.torch_dtype) == torch.bfloat16 and full_finetuning:
+        if _get_dtype(dtype_from_config(model.config)) == torch.bfloat16 and full_finetuning:
             # Use bfloat16 precision for full finetuning
             float32_mixed_precision = False
 
@@ -702,6 +954,7 @@ class FastBaseModel:
             train_embedding            = full_finetuning,
             train_lm_head              = full_finetuning,
             float32_mixed_precision    = float32_mixed_precision,
+            patch_modules_to_save      = True,
         )
 
         from transformers.trainer import Trainer
@@ -710,12 +963,12 @@ class FastBaseModel:
         pass
         patch_saving_functions(model, vision = True)
 
-        # Patch tokenizer to pad to the right
+        # Patch tokenizer to pad to the left
         m = model
         while hasattr(m, "model"):
             if hasattr(m, "_saved_temp_tokenizer"):
                 if hasattr(m._saved_temp_tokenizer, "tokenizer"):
-                    m._saved_temp_tokenizer.tokenizer.padding_side = "right"
+                    m._saved_temp_tokenizer.tokenizer.padding_side = "left"
             pass
             # Also set is_loaded_in_8bit to disable incorrect DDP
             m.is_loaded_in_8bit = True if not full_finetuning else False
@@ -723,7 +976,7 @@ class FastBaseModel:
         pass
         if hasattr(m, "_saved_temp_tokenizer"):
             if hasattr(m._saved_temp_tokenizer, "tokenizer"):
-                m._saved_temp_tokenizer.tokenizer.padding_side = "right"
+                m._saved_temp_tokenizer.tokenizer.padding_side = "left"
         pass
         # Also set is_loaded_in_8bit to disable incorrect DDP
         m.is_loaded_in_8bit = True if not full_finetuning else False
@@ -731,7 +984,7 @@ class FastBaseModel:
         # Clear deleted GPU items
         for _ in range(3):
             gc.collect()
-            if DEVICE_TYPE == "cuda":
+            if DEVICE_TYPE in ("cuda", "hip"):
                 torch.cuda.empty_cache()
             elif DEVICE_TYPE == "xpu":
                 torch.xpu.empty_cache()
@@ -739,6 +992,20 @@ class FastBaseModel:
         # Add for_inference and for_training
         model.for_training  = functools.partial(FastBaseModel.for_training,  model)
         model.for_inference = functools.partial(FastBaseModel.for_inference, model)
+        m = model
+        while hasattr(m, "model"):
+            m.for_training  = functools.partial(FastBaseModel.for_training,  m)
+            m.for_inference = functools.partial(FastBaseModel.for_inference, m)
+            m = m.model
+        # Set weight[padding_idx] = 0
+        with torch.no_grad():
+            for name, module in model.named_modules():
+                if type(module) is torch.nn.Embedding:
+                    if getattr(module, "weight", None) is not None and getattr(module, "padding_idx", None) is not None:
+                        if module.padding_idx < module.weight.shape[0]:
+                            module.weight[module.padding_idx] = 0
+        # Patch for torch.compiled inference
+        # FastBaseModel.pre_compile_for_inference(model_type, model, tokenizer)
         return model
     pass
 
@@ -761,6 +1028,7 @@ class FastBaseModel:
             _for_inference(m)
             m = m.model
         _for_inference(m)
+        model.eval() # to turn off training on modules deeper in
 
         # Since transformers 4.53, must turn off explicitly
         for module in model.modules():
@@ -805,13 +1073,19 @@ class FastBaseModel:
             # Pad tokenizer to the left
             if hasattr(m, "_saved_temp_tokenizer"): m._saved_temp_tokenizer.padding_side = "right"
             # Set a flag for generation!
-            if hasattr(m, "_flag_for_generation"): del m._flag_for_generation
+            if hasattr(m, "_flag_for_generation"):
+                try:
+                    # Weirdly sometimes cannot succeed so do a try except
+                    del m._flag_for_generation
+                except:
+                    pass
         pass
         m = model
         while hasattr(m, "model"):
             _for_training(m)
             m = m.model
         _for_training(m)
+        model.train() # to turn on training on modules deeper in
 
         # Since transformers 4.53, must turn on explicitly
         for module in model.modules():
